#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Generates 'predictions.csv', the submission file for the Fragile Families Challenge.

These files are generated from the data-cleaning notebook found in the following repository:
https://github.com/bjgoode/ffc-public.git

Specifics of the submission, including more details on the approach can be found there as well.

The original data, nor any derivative of the original data are not included in this package due to data release restrictions. Please contact the Fragile Families Challenge for more information:
http://www.fragilefamilieschallenge.org
"""

__author__ = "Brian J. Goode"

import pandas as pd

from sklearn.preprocessing import StandardScaler
from sklearn.feature_selection import SelectFromModel
from sklearn.linear_model import LassoCV, LogisticRegressionCV

import datetime

import numpy as np
import random as random

seed = 480
np.random.seed(seed)
random.seed(seed)

def scale_input(trainCol, features):
    Y = trainCol.dropna()
    idx = Y.index

    S = StandardScaler()
    idx_features = features.index
    FEATURES = S.fit_transform(features)
    X = FEATURES[idx-1,:]
    return FEATURES, X,Y, idx_features


def gpa():
    # Form Features
    gen_features = pd.read_csv('../output/imputing/.data/imputed_nodrops-v2.csv', index_col=0)
    
    mask = gen_features.columns[gen_features.columns.map(lambda x: x.startswith('k5')).tolist()]
    child_features = gen_features[mask]
    
    features = pd.concat([gen_features, child_features], axis=1)
    
    # Get Training Data
    df_train = pd.read_csv('../.data/train.csv',index_col=0)
    
    # Scale & Ready Inputs
    FEATURES, X_Train, Y_Train, idx_features = scale_input(df_train.gpa, features)
    
    # Get Important Features
    b = 500 #Can be set to length of data, but did this just to quickly verify model was somewhat in the ball park - see notebook.
    reg = LassoCV(cv=3, n_jobs=-1, selection='random')
    reg.fit(X_Train[:b],Y_Train[:b])
    
    ms = SelectFromModel(reg, threshold = .00001, prefit = True)
    x_red = ms.transform(X_Train)
    impt_features = features.columns[ms.get_support()]
    
    # Fit Reduced Features
    final_reg = LassoCV(selection='random', cv=10, n_jobs=-1)
    final_reg.fit(x_red,Y_Train)
    full_x_red = ms.transform(FEATURES)
    Y_Final = final_reg.predict(full_x_red)

    Y_Final = pd.DataFrame(Y_Final, index = idx_features, columns=['gpa'])
    return Y_Final


def grit():
    # Form Features
    gen_features = pd.read_csv('../output/imputing/.data/imputed_nodrops-v2.csv', index_col=0)
    specific_features = pd.read_csv('../output/imputing/.data/df_grit.csv', index_col=0)
    specific_features['total'] = specific_features.sum(axis=1)
    mask = gen_features.columns[gen_features.columns.map(lambda x: x.startswith('k5')).tolist()]
    child_features = gen_features[mask]
    features = pd.concat([gen_features,specific_features,child_features], axis = 1)

    # Get Training Data
    df_train = pd.read_csv('../.data/train.csv',index_col=0)
    
    # Scale & Ready Input
    FEATURES, X_Train, Y_Train, idx_features = scale_input(df_train.grit, features)
    
    # Get Important Features
    b = 500 #Can be set to length of data, but did this just to quickly verify model was somewhat in the ball park - see notebook.
    reg = LassoCV(cv=3, n_jobs=-1, selection='random')
    reg.fit(X_Train[:b],Y_Train[:b])
    
    ms = SelectFromModel(reg, threshold = .00001, prefit = True)
    x_red = ms.transform(X_Train)
    impt_features = features.columns[ms.get_support()]
        
    # Fit Reduced Features
    final_reg = LassoCV(selection='random', cv=10, n_jobs=-1)
    final_reg.fit(x_red,Y_Train)
    full_x_red = ms.transform(FEATURES)
    Y_Final = final_reg.predict(full_x_red)

    Y_Final = pd.DataFrame(Y_Final, index = idx_features, columns=['grit'])
    return Y_Final


def materialHardship():
    # Load Features
    specific_features = pd.read_csv('../output/imputing/.data/materialHardship_features.csv', index_col=0)
    specific_features_total = pd.read_csv('../output/imputing/.data/materialHardshipTotal_features.csv', index_col=0)
    features = specific_features.join([specific_features_total])
    
    # Get Training Data
    df_train = pd.read_csv('../.data/train.csv',index_col=0)
    
    # Scale & Ready Input
    FEATURES, X_Train, Y_Train, idx_features = scale_input(df_train.materialHardship, features)
        
    # Fit Reduced Features (note... features reduced by hand w/ input)
    final_reg = LassoCV(selection='random', cv=10, n_jobs=-1)
    final_reg.fit(X_Train,Y_Train)
    Y_Final = final_reg.predict(FEATURES)

    Y_Final = pd.DataFrame(Y_Final, index = idx_features, columns=['materialHardship'])
    return Y_Final
    

def eviction(MH_FEATURES):
    # Load Features
    specific_features = pd.read_csv('../output/imputing/.data/materialHardship_features.csv', index_col=0)
    specific_features_total = pd.read_csv('../output/imputing/.data/materialHardshipTotal_features.csv', index_col=0)
    specific_features_mth = MH_FEATURES#pd.read_csv('../output/solutions/Solution1/materialHardship.csv', index_col=0)
    specific_features = specific_features.join([specific_features_total])
    
    features = specific_features.join([specific_features_mth])
    
    # Get Training Data
    df_train = pd.read_csv('../.data/train.csv',index_col=0)
    
    # Scale & Ready Input
    FEATURES, X_Train, Y_Train, idx_features = scale_input(df_train.eviction, features)
    
    # Fit-predict (features reduced manually... see input)
    final_reg = LogisticRegressionCV(cv=10, n_jobs=-1)
    final_reg.fit(X_Train,Y_Train)
    Y_Final = final_reg.predict_proba(FEATURES)[:,1]
    
    Y_Final = pd.DataFrame(Y_Final, index = idx_features, columns=['eviction'])
    return Y_Final


def layoff():
    # Load Features
    specific_features = pd.read_csv('../output/imputing/.data/layoff_features.csv', index_col=0)
    specific_features['total'] = specific_features.sum(axis=1)
    
    features = specific_features
    
    # Get Training Data
    df_train = pd.read_csv('../.data/train.csv',index_col=0)
    
    # Scale & Ready Input
    FEATURES, X_Train, Y_Train, idx_features = scale_input(df_train.layoff, features)
    
    # Fit-predict (features reduced manually... see input)
    final_reg = LogisticRegressionCV(cv=10, n_jobs=-1)
    final_reg.fit(X_Train,Y_Train)
    Y_Final = final_reg.predict_proba(FEATURES)[:,1]
    
    Y_Final = pd.DataFrame(Y_Final, index = idx_features, columns=['layoff'])
    return Y_Final
    

def jobTraining():
    # Load Features
    specific_features = pd.read_csv('../output/imputing/.data/jobSkill_features.csv', index_col=0)
    specific_features['total'] = specific_features.sum(axis=1)
    
    features = specific_features
    
    # Get Training Data
    df_train = pd.read_csv('../.data/train.csv',index_col=0)
    
    # Scale & Ready Input
    FEATURES, X_Train, Y_Train, idx_features = scale_input(df_train.jobTraining, features)
    
    # Fit-predict (features reduced manually... see input)
    final_reg = LogisticRegressionCV(cv=10, n_jobs=-1)
    final_reg.fit(X_Train,Y_Train)
    Y_Final = final_reg.predict_proba(FEATURES)[:,1]
    
    Y_Final = pd.DataFrame(Y_Final, index = idx_features, columns=['jobTraining'])
    return Y_Final


def FormSubmissionFile():
    print('Evaluating GPA...')
    GPA = gpa()
    print('Evaluating GRIT...')
    GRIT = grit()
    print('Evaluating Material Hardship')
    MH = materialHardship()
    print('Evaluating Eviction')
    EVICT = eviction(MH)
    print('Evaluating Layoff')
    LAYOFF = layoff()
    print('Evaluating Job Training')
    JT = jobTraining()
    
    df = pd.DataFrame()
    for frame in [GPA, GRIT, MH, EVICT, LAYOFF, JT]:
        df = pd.concat([df,frame], axis=1)
    
    now = datetime.datetime.now().strftime("%Y-%m-%d_%H:%M")
    df.to_csv('../output/predictions/prediction-{}.csv'.format(now))
    
    return

if __name__ == "__main__":

    FormSubmissionFile()
    
